import pytest
import argparse
import datetime as dt
from klp import (
    visible,
    matches_python_expr,
)


@pytest.fixture
def mock_args():
    """Create a mock args object with default values."""
    args = argparse.Namespace()
    args.debug = False
    args.debug_where = False
    args.ts_key = None
    args.keys = []
    args.keys_not = []
    args.grep = []
    args.grep_not = []
    args.grep_by_key = {}
    args.grep_not_by_key = {}
    args.where = ""
    args.loglevels = None
    args.loglevels_not = None
    args.from_dt = None
    args.to_dt = None
    args.files = []  # Add default empty list of files
    args.files = []  # Added this line
    args.error_handling = "ignore"
    return args


@pytest.fixture
def sample_event():
    return {
        "timestamp": "2024-03-16T14:30:00Z",
        "level": "error",
        "message": "Test error message",
        "service": "api",
    }


def test_visible_with_grep(mock_args, sample_event):
    """Test grep filtering functionality."""
    import sys

    sys.modules["klp"].args = mock_args

    # Setup grep patterns
    import re

    mock_args.grep = [re.compile(r"error")]
    mock_args.grep_not = [re.compile(r"success")]

    # Should match error pattern
    assert visible(sample_event) is True

    # Should not match when error is removed
    sample_event["message"] = "Test success message"
    assert visible(sample_event) is False

    # Should not match when blocked pattern is present
    sample_event["message"] = "Error with success"
    assert visible(sample_event) is False


def test_visible_with_time_bounds(mock_args, sample_event):
    """Test time-based filtering."""
    import sys

    sys.modules["klp"].args = mock_args

    # Setup time bounds
    mock_args.from_dt = dt.datetime(2024, 3, 16, 14, 0, 0, tzinfo=dt.timezone.utc)
    mock_args.to_dt = dt.datetime(2024, 3, 16, 15, 0, 0, tzinfo=dt.timezone.utc)

    # Event within bounds
    assert visible(sample_event) is True

    # Event before lower bound
    sample_event["timestamp"] = "2024-03-16T13:59:59Z"
    assert visible(sample_event) is False

    # Event after upper bound
    sample_event["timestamp"] = "2024-03-16T15:00:01Z"
    # Add a second file to prevent StoppedEarly
    mock_args.files = ["file1.log", "file2.log"]
    assert visible(sample_event) is False

    # Invalid timestamp
    sample_event["timestamp"] = "invalid"
    assert visible(sample_event) is True  # Should not filter out invalid timestamps


def test_visible_with_log_levels(mock_args, sample_event):
    """Test log level filtering."""
    import sys

    sys.modules["klp"].args = mock_args

    # Include only specific levels
    mock_args.loglevels = ["error", "warning"]
    assert visible(sample_event) is True

    sample_event["level"] = "info"
    assert visible(sample_event) is False

    # Exclude specific levels
    mock_args.loglevels = None
    mock_args.loglevels_not = ["debug", "info"]
    sample_event["level"] = "error"
    assert visible(sample_event) is True

    sample_event["level"] = "info"
    assert visible(sample_event) is False


def test_visible_with_key_grep(mock_args, sample_event):
    """Test key-specific grep patterns."""
    import sys

    sys.modules["klp"].args = mock_args

    import re

    mock_args.grep_by_key = {"level": re.compile(r"error|warning")}
    assert visible(sample_event) is True

    sample_event["level"] = "info"
    assert visible(sample_event) is False

    # Test negative patterns
    mock_args.grep_by_key = {}
    mock_args.grep_not_by_key = {"service": re.compile(r"database")}
    assert visible(sample_event) is True

    sample_event["service"] = "database"
    assert visible(sample_event) is False


def test_matches_python_expr(mock_args):
    """Test Python expression matching against events."""
    # Set the args directly in the klp.globals() namespace
    import sys

    sys.modules["klp"].args = mock_args

    event = {
        "level": "ERROR",
        "status": 500,
        "message": "Test failed",
        "timestamp": "2024-03-16T14:30:00Z",
    }

    # Basic comparisons
    assert matches_python_expr("level == 'ERROR'", event) is True
    assert matches_python_expr("status >= 500", event) is True
    assert matches_python_expr("'failed' in message", event) is True

    # Complex expressions
    assert matches_python_expr("level == 'ERROR' and status >= 500", event) is True
    assert matches_python_expr("len(message) > 5 and 'Test' in message", event) is True

    # Using imported functions
    assert matches_python_expr("'2024' in timestamp", event) is True

    # Invalid expressions
    assert matches_python_expr("invalid_variable", event) is False
    assert matches_python_expr("nonexistent.function()", event) is False



import pytest
import argparse
import datetime as dt
from klp import guess_datetime, timedelta_from, add_ts_delta


@pytest.fixture
def mock_args():
    """Create a mock args object with default values."""
    args = argparse.Namespace()
    args.debug = False
    args.localtime = False
    args.utc = True
    args.ts_key = None
    args.error_handling = "ignore"
    return args


def test_timedelta_from_valid():
    """Test parsing of valid time intervals."""
    assert timedelta_from("1h") == dt.timedelta(hours=1)
    assert timedelta_from("30m") == dt.timedelta(minutes=30)
    assert timedelta_from("1.5h") == dt.timedelta(hours=1, minutes=30)
    assert timedelta_from("90s") == dt.timedelta(seconds=90)


def test_timedelta_from_invalid():
    """Test parsing of invalid time intervals."""
    with pytest.raises(argparse.ArgumentTypeError, match="Invalid timespec: invalid"):
        timedelta_from("invalid")

    with pytest.raises(
        argparse.ArgumentTypeError, match="Durations cannot be negative"
    ):
        timedelta_from("-1h")


def test_add_ts_delta(mock_args):
    """Test timestamp delta calculations."""
    import sys

    sys.modules["klp"].args = mock_args

    event1 = {"timestamp": "2024-03-16T14:30:00Z"}
    event2 = {"timestamp": "2024-03-16T14:30:01Z"}
    event3 = {"timestamp": "2024-03-16T14:31:00Z"}

    # First event should get delta from epoch
    result1, last_ts = add_ts_delta(event1, None)
    assert "_klp_timedelta" in result1

    # One second difference
    result2, last_ts = add_ts_delta(event2, last_ts)
    assert result2["_klp_timedelta"] == "0:00:01.000000"

    # One minute difference
    result3, last_ts = add_ts_delta(event3, last_ts)
    assert result3["_klp_timedelta"] == "0:00:59.000000"

    # Invalid timestamp
    invalid_event = {"timestamp": "invalid"}
    result_invalid, last_ts = add_ts_delta(invalid_event, last_ts)
    assert "_klp_timedelta" not in result_invalid


def test_guess_datetime_military_ns():
    result = guess_datetime("2022-12-30T20:55:00.123000000Z")
    expected = dt.datetime(2022, 12, 30, 20, 55, 0, 123000, tzinfo=dt.timezone.utc)
    assert result == expected


def test_guess_datetime_military_ms():
    result = guess_datetime("2022-12-30T20:55:00.123Z")
    expected = dt.datetime(2022, 12, 30, 20, 55, 0, 123000, tzinfo=dt.timezone.utc)
    assert result == expected


def test_guess_datetime_military_s():
    result = guess_datetime("2022-12-30T20:55:00Z")
    expected = dt.datetime(2022, 12, 30, 20, 55, 0, 0, tzinfo=dt.timezone.utc)
    assert result == expected


def test_guess_datetime_offset():
    result = guess_datetime("2022-12-30T22:55:00+02:00")
    expected = dt.datetime(2022, 12, 30, 20, 55, 0, 0, tzinfo=dt.timezone.utc)
    assert result == expected


def test_timedelta_from_microseconds():
    result = timedelta_from("4us")
    expected = dt.timedelta(microseconds=4)
    assert result == expected


def test_timedelta_from_milliseconds():
    result = timedelta_from("4ms")
    expected = dt.timedelta(microseconds=4000)
    assert result == expected


def test_timedelta_from_seconds():
    result = timedelta_from("54s")
    expected = dt.timedelta(seconds=54)
    assert result == expected


def test_timedelta_from_minutes():
    result = timedelta_from("4m")
    expected = dt.timedelta(minutes=4)
    assert result == expected


def test_timedelta_from_hours():
    result = timedelta_from("4h")
    expected = dt.timedelta(hours=4)
    assert result == expected


def test_timedelta_from_days():
    result = timedelta_from("2d")
    expected = dt.timedelta(days=2)
    assert result == expected


def test_timedelta_from_weeks():
    result = timedelta_from("5w")
    expected = dt.timedelta(weeks=5)
    assert result == expected


def test_timedelta_from_floatseconds():
    result = timedelta_from("1.567s")
    expected = dt.timedelta(seconds=1, microseconds=567000)
    assert result == expected


def test_timedelta_from_no_unit():
    with pytest.raises(argparse.ArgumentTypeError):
        timedelta_from("567")


def test_timedelta_from_no_negatives():
    with pytest.raises(argparse.ArgumentTypeError):
        timedelta_from("-3s")


def test_guess_datetime_formats():
    """Test datetime parsing from various formats."""

    # Convert to UTC for consistent comparison
    def to_utc(dt_obj):
        if dt_obj is None:
            return None
        return dt_obj.astimezone(dt.timezone.utc)

    test_cases = [
        # ISO format with timezone
        (
            "2024-03-16T14:30:00Z",
            dt.datetime(2024, 3, 16, 14, 30, tzinfo=dt.timezone.utc),
        ),
        (
            "2024-03-16T14:30:00+00:00",
            dt.datetime(2024, 3, 16, 14, 30, tzinfo=dt.timezone.utc),
        ),
        # Dates with timezone info
        (
            "2024-03-16 14:30:00Z",
            dt.datetime(2024, 3, 16, 14, 30, tzinfo=dt.timezone.utc),
        ),
        (
            "2024-03-16 14:30:00+00:00",
            dt.datetime(2024, 3, 16, 14, 30, tzinfo=dt.timezone.utc),
        ),
        # Date only formats (these will be interpreted in local time)
        (
            "2024-03-16",
            dt.datetime(2024, 3, 16, 0, 0).astimezone().astimezone(dt.timezone.utc),
        ),
        (
            "2024-03",
            dt.datetime(2024, 3, 1, 0, 0).astimezone().astimezone(dt.timezone.utc),
        ),
        # Common Log Format
        (
            "16/Mar/2024:14:30:00 +0000",
            dt.datetime(2024, 3, 16, 14, 30, tzinfo=dt.timezone.utc),
        ),
        # Invalid formats
        ("invalid", None),
        ("", None),
    ]

    for input_str, expected in test_cases:
        result = guess_datetime(input_str)
        if expected is None:
            assert result is None, f"Expected None for input {input_str}, got {result}"
        else:
            result = to_utc(result)
            assert abs(result - expected) < dt.timedelta(
                seconds=1
            ), f"Failed for input {input_str}: got {result}, expected {expected}"



import os
import pytest
import subprocess


# Helper to get paths relative to repo root
def get_repo_path(*paths):
    """Get absolute path relative to current test directory."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", *paths))


KLP_PATH = get_repo_path("klp.py")


@pytest.fixture
def sample_logfile(tmp_path):
    """Create a log file with predictable timestamps and levels."""
    logfile = tmp_path / "test.log"
    with open(logfile, "w") as f:
        # Normal sequence
        f.write('time="2024-03-16T10:00:00Z" level=info message="Starting service"\n')
        f.write(
            'time="2024-03-16T10:00:01Z" level=debug message="Initializing cache"\n'
        )
        f.write('time="2024-03-16T10:00:02Z" level=warn message="Cache slow"\n')
        # Gap of 5 minutes
        f.write('time="2024-03-16T10:05:00Z" level=error message="Connection lost"\n')
        f.write('time="2024-03-16T10:05:01Z" level=info message="Reconnecting"\n')
        # Rapid sequence
        f.write('time="2024-03-16T10:05:01.100Z" level=debug message="Attempt 1"\n')
        f.write('time="2024-03-16T10:05:01.200Z" level=debug message="Attempt 2"\n')
        f.write('time="2024-03-16T10:05:01.300Z" level=debug message="Attempt 3"\n')
        # Success after attempts
        f.write('time="2024-03-16T10:05:02Z" level=info message="Connected"\n')
    return str(logfile)


def run_klp(logfile, *args, color=False):
    """Run klp with given args and return output."""
    cmd = ["python", KLP_PATH]
    if not color:
        cmd.append("--no-color")  # Ensure consistent output for testing
    cmd.extend(args)
    cmd.append(logfile)

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0
    return result.stdout, result.stderr


def test_levelmap(sample_logfile):
    """Test level map visualization."""
    stdout, stderr = run_klp(sample_logfile, "--levelmap")

    # Should show timestamp followed by level indicators
    # i(nfo) d(ebug) w(arn) e(rror) i(nfo) d d d i
    lines = stdout.splitlines()
    assert len(lines) == 1  # All in one line

    # Extract the levelmap part (after the timestamp)
    levelmap = lines[0].split(" ", 1)[1].strip()
    assert levelmap == "idweidddi"


def test_mark_gaps(sample_logfile):
    """Test gap marking between events."""
    stdout, stderr = run_klp(sample_logfile, "--mark-gaps=1m")
    stderr_lines = stderr.splitlines()

    # Find the gap marker line
    gap_lines = [line for line in stderr_lines if "time gap:" in line]
    assert len(gap_lines) > 0

    # The first gap should be around 4:58
    gap_text = gap_lines[0]
    assert "time gap: 0:04:58" in gap_text


def test_fuse(sample_logfile):
    """Test fusing of rapid events."""
    stdout, stderr = run_klp(sample_logfile, "--fuse=0.5s")
    lines = stdout.splitlines()

    # Find the rapid sequence
    attempt_lines = [l for l in lines if "Attempt" in l]
    # Should only show last attempt with prefix
    assert len(attempt_lines) == 1
    assert attempt_lines[0].startswith("5 ")
    assert "Attempt 3" in attempt_lines[0]


def test_fuse_with_gaps(sample_logfile):
    """Test fuse with gap marking."""
    stdout, stderr = run_klp(sample_logfile, "--fuse=0.5s", "--mark-gaps=1m")
    stdout_lines = stdout.splitlines()
    stderr_lines = stderr.splitlines()

    # Verify both gaps and fusing
    gap_lines = [l for l in stderr_lines if "time gap:" in l]
    fused_lines = [l for l in stdout_lines if l.startswith("5 ")]

    assert len(gap_lines) > 0  # Should have gap markers
    assert len(fused_lines) > 0  # Should have fused events


def test_each_key(sample_logfile):
    """Test printing each key on a separate line."""
    stdout, stderr = run_klp(sample_logfile, "--each-key")
    lines = stdout.splitlines()

    # First event should be split into 3 lines (time, level, message)
    first_event = lines[:3]
    assert any("time=" in line for line in first_event)
    assert any("level=" in line for line in first_event)
    assert any("message=" in line for line in first_event)


def test_context_prefix(sample_logfile):
    """Test context prefixes for grep matches."""
    stdout, stderr = run_klp(
        sample_logfile, "--grep=error", "--before-context=1", "--after-context=1"
    )
    lines = stdout.splitlines()

    # Should show before, match, and after lines with prefixes
    assert any(line.startswith("/ ") for line in lines)  # Before
    assert any(line.startswith("* ") for line in lines)  # Match
    assert any(line.startswith("\\ ") for line in lines)  # After


def test_stats_output(sample_logfile):
    """Test stats output format."""
    stdout, stderr = run_klp(sample_logfile, "--stats")

    # Stats should go to stderr
    assert "Events shown:" in stderr
    assert "Time span shown:" in stderr
    assert "Keys seen:" in stderr
    assert "Log levels seen:" in stderr

    # Should list all our log levels
    assert all(level in stderr for level in ["info", "debug", "warn", "error"])



from klp import (
    parse_logfmt,
    parse_jsonl,
    parse_clf,
    parse_combined,
    parse_unix,
)


def test_parse_logfmt():
    """Test parsing of logfmt formatted strings."""
    text = 'timestamp="2024-03-16T14:30:00Z" level=info msg="Starting service"'
    expected = {
        "timestamp": "2024-03-16T14:30:00Z",
        "level": "info",
        "msg": "Starting service",
    }
    assert parse_logfmt(text) == expected


def test_parse_jsonl():
    """Test parsing of JSONL formatted strings."""
    text = '{"timestamp": "2024-03-16T14:30:00Z", "level": "info", "msg": "Test"}'
    expected = {"timestamp": "2024-03-16T14:30:00Z", "level": "info", "msg": "Test"}
    assert parse_jsonl(text) == expected


def test_parse_clf():
    """Test Common Log Format parsing."""
    # Standard line
    line = '127.0.0.1 - frank [16/Mar/2024:14:30:00 +0000] "GET /index.html HTTP/1.1" 200 1234'
    result = parse_clf(line)
    assert result["host"] == "127.0.0.1"
    assert result["ident"] == "-"
    assert result["user"] == "frank"
    assert result["time"] == "16/Mar/2024:14:30:00 +0000"
    assert result["request"] == "GET /index.html HTTP/1.1"
    assert result["status"] == "200"
    assert result["size"] == "1234"

    # With missing fields using "-"
    line = '127.0.0.1 - - [16/Mar/2024:14:30:00 +0000] "GET /index.html HTTP/1.1" 200 -'
    result = parse_clf(line)
    assert result["size"] == "0"  # Dash is converted to 0

    # Invalid format
    assert parse_clf("Invalid line") == {}


def test_parse_combined():
    """Test Combined Log Format parsing with various cases."""
    # Looking at RE_COMBINED in klp.py, the regex matches these field names

    # Standard line
    line = '127.0.0.1 - frank [16/Mar/2024:14:30:00 +0000] "GET /index.html HTTP/1.1" 200 1234 "http://example.com" "Mozilla/5.0"'
    result = parse_combined(line)
    assert result["host"] == "127.0.0.1"
    assert result["ident"] == "-"
    assert result["user"] == "frank"
    assert result["time"] == "16/Mar/2024:14:30:00 +0000"
    assert result["request"] == "GET /index.html HTTP/1.1"
    assert result["status"] == "200"
    assert result["size"] == "1234"
    assert result["referrer"] == "http://example.com"
    assert result["agent"] == "Mozilla/5.0"

    # With missing fields using "-"
    line = '127.0.0.1 - - [16/Mar/2024:14:30:00 +0000] "GET /index.html HTTP/1.1" 200 - "-" "-"'
    result = parse_combined(line)
    assert result["ident"] == "-"
    assert result["user"] == "-"
    assert result["size"] == "0"  # Dash is converted to 0
    assert result["referrer"] == "-"
    assert result["agent"] == "-"

    # Invalid format
    assert parse_combined("Invalid line") == {}


def test_parse_unix():
    """Test Unix syslog format parsing."""
    # Standard format with PID
    line = "Mar 16 14:30:00 localhost sshd[12345]: Connection accepted"
    result = parse_unix(line)
    assert result["timestamp"] == "Mar 16 14:30:00"
    assert result["hostname"] == "localhost"
    assert result["service"] == "sshd"
    assert result["pid"] == "12345"
    assert result["message"] == "Connection accepted"

    # Without PID
    line = "Mar 16 14:30:00 localhost sshd: Connection accepted"
    result = parse_unix(line)
    assert "pid" not in result
    assert result["service"] == "sshd"
    assert result["message"] == "Connection accepted"

    # Invalid format
    assert parse_unix("Invalid line") == {}


def test_parse_logfmt_basic():
    text = "key1=value1 key2=value2"
    expected = {"key1": "value1", "key2": "value2"}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_with_spaces():
    text = 'key1="value with spaces" key2=value2'
    expected = {"key1": "value with spaces", "key2": "value2"}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_with_escaped_quotes():
    text = 'key1="value with \\"escaped quotes\\"" key2=value2'
    expected = {"key1": 'value with "escaped quotes"', "key2": "value2"}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_with_quotes_inside():
    text = 'key1=singlequote\'s key2=double"quot"es'
    expected = {"key1": "singlequote's", "key2": 'double"quot"es'}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_with_unicode():
    text = 'schlüssel=Schloß kauf="Äpfel aus Köln"'
    expected = {"schlüssel": "Schloß", "kauf": "Äpfel aus Köln"}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_with_dots_in_key():
    text = "my.long.key=value1 key2=value2"
    expected = {"my.long.key": "value1", "key2": "value2"}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_with_equal_in_val():
    text = "key=2+2=4 key2=test"
    expected = {"key": "2+2=4", "key2": "test"}
    assert parse_logfmt(text) == expected


def test_parse_logfmt_empty_string():
    assert parse_logfmt("") == {}



import subprocess
import json
import os
import pytest

# Helper to get paths relative to git repo root
def get_repo_path(*paths):
    """Get absolute path relative to current test directory."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..', *paths))

# Get path to klp.py in repo root
KLP_PATH = get_repo_path('klp.py')

@pytest.fixture
def temp_logfile(tmp_path):
    """Create a temporary logfile with test data."""
    logfile = tmp_path / "test.log"
    with open(logfile, "w") as f:
        f.write('timestamp="2024-03-16T10:00:00Z" level=info message="Test message 1"\n')
        f.write('timestamp="2024-03-16T10:01:00Z" level=error message="Test error"\n')
        f.write('timestamp="2024-03-16T10:02:00Z" level=info message="Test message 2"\n')
    return str(logfile)

@pytest.fixture
def temp_jsonfile(tmp_path):
    """Create a temporary JSON file with test data."""
    jsonfile = tmp_path / "test.json"
    data = [
        {"timestamp": "2024-03-16T10:00:00Z", "level": "info", "message": "Test message"},
        {"timestamp": "2024-03-16T10:01:00Z", "level": "error", "message": "Test error"}
    ]
    with open(jsonfile, "w") as f:
        json.dump(data, f)
    return str(jsonfile)

@pytest.fixture
def temp_csvfile(tmp_path):
    """Create a temporary CSV file with test data."""
    csvfile = tmp_path / "test.csv"
    with open(csvfile, "w") as f:
        f.write("timestamp,level,message\n")
        f.write("2024-03-16T10:00:00Z,info,Test message\n")
        f.write("2024-03-16T10:01:00Z,error,Test error\n")
    return str(csvfile)

def test_basic_logfmt_processing(temp_logfile):
    """Test basic logfmt file processing."""
    cmd = ["python", KLP_PATH, "--output-format=jsonl", temp_logfile]
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0, f"Process failed with error: {result.stderr}"

    # Parse output lines as JSON
    outputs = [json.loads(line) for line in result.stdout.strip().split("\n")]
    assert len(outputs) == 3
    assert outputs[0]["level"] == "info"
    assert outputs[1]["level"] == "error"
    assert outputs[2]["level"] == "info"

def test_json_processing(temp_jsonfile):
    """Test JSON file processing."""
    cmd = [
        "python",
        KLP_PATH,
        "--input-format=json",
        "--output-format=jsonl",
        temp_jsonfile,
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0, f"Process failed with error: {result.stderr}"

    outputs = [json.loads(line) for line in result.stdout.strip().split("\n")]
    assert len(outputs) == 2
    assert outputs[0]["level"] == "info"
    assert outputs[1]["level"] == "error"

def test_csv_processing(temp_csvfile):
    """Test CSV file processing."""
    cmd = [
        "python",
        KLP_PATH,
        "--input-format=csv",
        "--output-format=jsonl",
        temp_csvfile,
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0, f"Process failed with error: {result.stderr}"

    outputs = [json.loads(line) for line in result.stdout.strip().split("\n")]
    assert len(outputs) == 2
    assert "timestamp" in outputs[0]
    assert "level" in outputs[0]
    assert "message" in outputs[0]

def test_grep_functionality(temp_logfile):
    """Test grep functionality."""
    cmd = [
        "python",
        KLP_PATH,
        "--output-format=jsonl",
        "--grep=error",
        temp_logfile,
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0, f"Process failed with error: {result.stderr}"

    outputs = [json.loads(line) for line in result.stdout.strip().split("\n")]
    assert len(outputs) == 1
    assert outputs[0]["level"] == "error"

def test_timestamp_filtering(temp_logfile):
    """Test timestamp-based filtering."""
    cmd = [
        "python",
        KLP_PATH,
        "--output-format=jsonl",
        "--from=2024-03-16T10:01:00Z",
        temp_logfile,
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0, f"Process failed with error: {result.stderr}"

    outputs = [json.loads(line) for line in result.stdout.strip().split("\n")]
    assert len(outputs) == 2
    assert outputs[0]["timestamp"] >= "2024-03-16T10:01:00Z"



import os
import json
import subprocess
import time
from datetime import datetime


# Helper to get paths relative to repo root
def get_repo_path(*paths):
    """Get absolute path relative to current test directory."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", *paths))


KLP_PATH = get_repo_path("klp.py")
GENERATE_STREAM_PATH = get_repo_path("tools/generate_stream.py")


def run_pipeline(generator_args=None, klp_args=None, duration=0.2):
    """
    Run a pipeline of generate_stream.py | klp.py with given arguments.

    Args:
        generator_args: List of arguments for generate_stream.py
        klp_args: List of arguments for klp.py
        duration: How long to let the pipeline run (seconds)

    Returns:
        tuple: (stdout lines, stderr lines)
    """
    # Ensure scripts are executable
    os.chmod(GENERATE_STREAM_PATH, 0o755)
    os.chmod(KLP_PATH, 0o755)

    # Start generator process
    generator_cmd = [GENERATE_STREAM_PATH]
    if generator_args:
        generator_cmd.extend(generator_args)

    generator = subprocess.Popen(
        generator_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )

    # Start klp process
    klp_cmd = [KLP_PATH]
    if klp_args:
        klp_cmd.extend(klp_args)

    klp = subprocess.Popen(
        klp_cmd,
        stdin=generator.stdout,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    # Add these lines after starting klp process
    generator.stdout.flush()  # Force flush the generator output
    time.sleep(0.1)  # Give processes time to start up

    # Let it run
    time.sleep(duration)

    # Add a small delay before cleanup
    time.sleep(0.1)

    # Clean kill of generator will cause klp to exit when stdin closes
    generator.terminate()
    generator.wait()
    stdout, stderr = klp.communicate(timeout=2)

    return stdout.splitlines(), stderr.splitlines()


def test_basic_streaming():
    """Test basic log streaming functionality."""
    stdout, stderr = run_pipeline(duration=0.2)

    # Should get roughly 0.2 seconds worth of logs (1 per second)
    assert 1 <= len(stdout) <= 3  # Allow some variance

    # Each line should be properly formatted
    for line in stdout:
        # Basic format checks
        assert "time=" in line
        assert "level=" in line
        assert "message=" in line


def test_streaming_jsonl_output():
    """Test JSON Lines output format."""
    stdout, stderr = run_pipeline(klp_args=["--output-format=jsonl"], duration=0.2)

    # Verify each line is valid JSON
    for line in stdout:
        entry = json.loads(line)
        assert isinstance(entry, dict)
        assert "time" in entry
        assert "level" in entry
        assert "message" in entry

        # Verify timestamp format
        time_str = entry["time"].rstrip("Z")
        assert datetime.fromisoformat(time_str)  # Should not raise exception


def test_streaming_with_grep():
    """Test grepping from stream."""
    stdout, stderr = run_pipeline(
        klp_args=["--grep=error", "--output-format=jsonl"],
        duration=0.4,  # Longer duration to ensure we get some errors
    )

    # Every line should contain 'error'
    for line in stdout:
        entry = json.loads(line)
        assert "error" in line.lower()
        # Should be in level field specifically
        assert entry["level"] == "error"


def test_streaming_with_key_selection():
    """Test selecting specific keys from stream."""
    stdout, stderr = run_pipeline(
        klp_args=["-k", "time,level,message", "--output-format=jsonl"], duration=0.2
    )

    # Verify only requested fields are present
    for line in stdout:
        entry = json.loads(line)
        assert set(entry.keys()) == {"time", "level", "message"}


def test_streaming_stats():
    """Test statistics generation from stream."""
    stdout, stderr = run_pipeline(klp_args=["--stats-only"], duration=0.3)

    # Stats should be written to stderr
    assert len(stderr) > 0
    stats_text = "\n".join(stderr)

    # Basic stats checks
    assert "Events shown:" in stats_text
    assert "Time span shown:" in stats_text
    assert "Keys seen:" in stats_text
    assert "Log levels seen:" in stats_text


def test_streaming_levelmap():
    """Test level map visualization."""
    stdout, stderr = run_pipeline(klp_args=["--levelmap"], duration=0.3)

    # Should get a character map of log levels
    assert len(stdout) > 0
    levelmap = "".join(stdout)

    # Should contain first letters of log levels
    assert any(c in levelmap for c in "diwef")  # debug, info, warn, error, fatal


def test_streaming_with_time_window():
    """Test time windowed streaming."""
    stdout, stderr = run_pipeline(
        klp_args=["--output-format=jsonl", "--duration=1s"], duration=0.3
    )

    if stdout:  # If we got any output
        # Parse timestamps
        times = []
        for line in stdout:
            entry = json.loads(line)
            time_str = entry["time"].rstrip("Z")
            times.append(datetime.fromisoformat(time_str))

        # Verify time window
        if len(times) >= 2:
            time_span = max(times) - min(times)
            assert time_span.total_seconds() <= 1.5  # Allow some slack


def test_error_handling():
    """Test handling of malformed input."""
    # Run klp with invalid input
    process = subprocess.Popen(
        [KLP_PATH],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    # Send some invalid logfmt
    stdout, stderr = process.communicate("this is not valid logfmt\n")

    # Should handle invalid input gracefully
    assert process.returncode == 0  # Shouldn't crash



import pytest
import argparse
from klp import flatten_object, reorder


@pytest.fixture
def mock_args():
    """Create a mock args object with default values."""
    args = argparse.Namespace()
    args.debug = False
    args.keys = []
    args.keys_not = []
    return args


def test_flatten_object():
    """Test JSON object flattening with nested structures."""
    # Nested dictionaries
    nested = {"a": {"b": 1, "c": {"d": 2}}, "e": 3}
    assert flatten_object(nested) == {"a.b": 1, "a.c.d": 2, "e": 3}

    # Arrays
    with_arrays = {"a": [{"b": 1}, {"c": 2}], "d": 3}
    assert flatten_object(with_arrays) == {"a.0.b": 1, "a.1.c": 2, "d": 3}

    # Mixed nested structures
    mixed = {"a": [{"b": {"c": 1}}, 2], "d": {"e": [3, 4]}}
    assert flatten_object(mixed) == {"a.0.b.c": 1, "a.1": 2, "d.e.0": 3, "d.e.1": 4}


def test_reorder(mock_args):
    """Test event key reordering."""
    import sys

    event = {
        "timestamp": "2024-03-16T14:30:00Z",
        "level": "ERROR",
        "message": "Test message",
        "extra": "data",
    }

    # Test with specific keys
    mock_args.keys = ["message", "level"]
    mock_args.keys_not = []
    sys.modules["klp"].args = mock_args
    result = reorder(event)
    assert list(result.keys()) == ["message", "level"]

    # Test with excluded keys
    mock_args.keys = []
    mock_args.keys_not = ["extra"]
    sys.modules["klp"].args = mock_args
    result = reorder(event)
    assert "extra" not in result
    assert all(k in result for k in ["timestamp", "level", "message"])

    # Test with no specifications
    mock_args.keys = []
    mock_args.keys_not = []
    sys.modules["klp"].args = mock_args
    result = reorder(event)
    assert result == event



import pytest
import re
from klp import (
    BUILTIN_REGEXES,
    EStr,
    escape_doublequotes_quoted,
    extract_blocks,
    extract_json,
    make_greppable,
    parse_kv,
    sanitize_key,
    split_startswith,
    unescape,
)


def test_parse_kv_complex():
    """Test key-value parsing with different separators and formats."""
    # Basic case
    assert parse_kv("key1=val1 key2=val2") == {"key1": "val1", "key2": "val2"}

    # Custom separators
    assert parse_kv("key1:val1;key2:val2", sep=";", kvsep=":") == {
        "key1": "val1",
        "key2": "val2",
    }

    # Handle missing values
    assert parse_kv("key1=val1 key2= key3=val3") == {
        "key1": "val1",
        "key2": "",
        "key3": "val3",
    }

    # Handle equals in values
    assert parse_kv("key1=val=1 key2=val2") == {"key1": "val=1", "key2": "val2"}

    # Note: parse_kv splits on whitespace by default
    assert parse_kv("key1=val1 value key2=val2") == {"key1": "val1", "key2": "val2"}


def test_split_startswith():
    """Test splitting strings on regex patterns."""
    # Basic split on newlines
    text = "First\\nSecond\\nThird"
    assert split_startswith(text, r"\\n") == ["First", "\\nSecond", "\\nThird"]

    # Empty pattern
    assert split_startswith("text", "") == ["text"]

    # No matches
    assert split_startswith("abc", "x") == ["abc"]

    # Multiple consecutive matches
    assert split_startswith("axbxcx", "x") == ["a", "xb", "xc", "x"]

    # Empty string
    assert split_startswith("", "x") == [""]


def test_extract_json():
    """Test JSON extraction from text."""
    # Basic JSON object
    text = 'before {"key": "value"} after'
    assert extract_json(text) == '{"key": "value"}'

    # Test invalid JSON
    with pytest.raises(ValueError):
        extract_json("invalid")


def test_make_greppable():
    """Test creation of greppable string from event dict."""
    event = {"key1": "value1", "key2": "value2"}
    expected = 'key1="value1" key2="value2"'
    assert make_greppable(event) == expected


def test_sanitize_key():
    """Test sanitization of keys."""
    assert sanitize_key("user.name@domain") == "user_name_domain"
    assert sanitize_key("valid_key_123") == "valid_key_123"


def test_builtin_regexes_compile():
    """Test that all builtin regex patterns compile successfully."""
    for name, regexes in BUILTIN_REGEXES.items():
        for regex in regexes:
            try:
                re.compile(regex)
            except re.error as e:
                pytest.fail(f"Regex compilation for {name} failed with error: {e.msg}")


def test_escape_doublequotes():
    """Test escaping text with double quotes."""
    text = 'a "string" with double quotes'
    expected = 'a \\"string\\" with double quotes'
    assert escape_doublequotes_quoted(text) == expected


def test_escape_single_quotes():
    """Test escaping text with single quotes."""
    text = "a 'string' with single quotes"
    expected = "a 'string' with single quotes"
    assert escape_doublequotes_quoted(text) == expected


def test_escape_sequences():
    """Test escaping special characters."""
    text = "Escape seqs: \x00\t\n"
    expected = "Escape seqs: \\x00\\t\\n"
    assert escape_doublequotes_quoted(text) == expected


def test_unescape():
    """Test unescaping various escaped sequences."""
    assert unescape(r"This is a \"test\" string") == 'This is a "test" string'
    assert unescape(r"Escape sequences: \n \t") == "Escape sequences: \n \t"
    assert unescape("значение со spaces") == "значение со spaces"
    assert unescape("\\x01") == "\x01"


def test_extract_json_valid_object():
    """Test extracting valid JSON objects."""
    assert extract_json('{"name": "John", "age": 30}') == '{"name": "John", "age": 30}'
    assert extract_json("[1, 2, 3, 4]") == "[1, 2, 3, 4]"
    assert (
        extract_json('{"person": {"name": "John", "age": 30}, "city": "New York"}')
        == '{"person": {"name": "John", "age": 30}, "city": "New York"}'
    )


def test_extract_json_with_surrounding_text():
    """Test extracting JSON from text with non-JSON content."""
    assert (
        extract_json('Hello world {"name": "John", "age": 30}')
        == '{"name": "John", "age": 30}'
    )
    assert (
        extract_json('{"name": "John", "age": 30} and more text')
        == '{"name": "John", "age": 30}'
    )


def test_extract_json_invalid():
    """Test extracting invalid JSON raises ValueError."""
    with pytest.raises(ValueError):
        extract_json('{"name": "John" "age": 30}')
    with pytest.raises(ValueError):
        extract_json("Just plain text")
    with pytest.raises(ValueError):
        extract_json("")


def test_extract_json_multiple():
    """Test extracting first JSON object when multiple are present."""
    assert extract_json('{"name": "John"} {"age": 30}') == '{"name": "John"}'


@pytest.mark.parametrize(
    "input_lines, start_after, start_with, stop_before, stop_with, expected_blocks",
    [
        # Basic block extraction
        (
            ["start", "content1", "content2", "end"],
            None,
            ["start"],
            None,
            ["end"],
            [(["start", "content1", "content2", "end"], 1, 4)],
        ),
        # Multiple blocks
        (
            ["start", "c1", "end", "start", "c2", "end"],
            None,
            ["start"],
            None,
            ["end"],
            [(["start", "c1", "end"], 1, 3), (["start", "c2", "end"], 4, 6)],
        ),
        # Start after pattern
        (
            ["before", "start", "content", "end"],
            ["before"],
            None,
            None,
            ["end"],
            [(["start", "content", "end"], 2, 4)],
        ),
    ],
)
def test_extract_blocks(
    input_lines, start_after, start_with, stop_before, stop_with, expected_blocks
):
    """Test block extraction with different patterns."""
    blocks = list(
        extract_blocks(
            iter(input_lines),
            start_after=start_after,
            start_with=start_with,
            stop_before=stop_before,
            stop_with=stop_with,
        )
    )
    assert blocks == expected_blocks


def test_estr_init():
    s = EStr("This is a test")
    assert s == "This is a test"


def test_estr_init_empty():
    s = EStr("")
    assert s == ""


def test_estr_basic_ops():
    s = EStr("This is a test")
    assert str(s) == "This is a test"
    assert len(s) == 14
    assert s[0] == "T"
    assert s[1] == "h"
    assert s[2] == "i"


def test_estr_col():
    s = EStr("This is a test")
    assert s.col(0) == "This"
    assert s.col(1) == "is"
    assert s.col(2) == "a"
    assert s.col(3) == "test"
    assert s.col(4) == None


def test_estr_cols_basic():
    s = EStr("This is  a test  with 7 columns")
    assert s.cols() == ["This", "is", "a", "test", "with", "7", "columns"]
    assert s.cols(0, 3) == ["This", "test"]
    assert s.cols(0, -1, 2, 2) == ["This", "columns", "a", "a"]
    assert s.cols("0,3") == "This test"
    assert s.cols("1") == "is"
    assert s.cols(1) == "is"
    assert s.cols("14") == ""
    assert s.cols("1:3") == "is a"
    assert s.cols("-2,2,4:") == "7 a with 7 columns"


def test_estr_cols_with_separators():
    s = EStr("This|is a|test with|4 columns")
    assert s.cols("1:3", sep="|") == "is a test with"
    assert s.cols("-2,2,4:", sep="|", outsep=":") == "test with:test with"


def test_estr_cols_empty_sep():
    # Character-by-character splitting with empty separators
    text = EStr("Hello")
    assert text.cols("0,2,4", sep="", outsep="") == "Hlo"  # Single arg joins indices
    assert text.cols("1:4", sep="", outsep="") == "ell"  # Single slice
    assert text.cols("0,2", "4", sep="", outsep="") == [
        "Hl",
        "o",
    ]  # Multiple args as list
    assert text.cols(0, "2:4", sep="", outsep="") == [
        "H",
        "ll",
    ]  # Mix of single index and slice

    assert text.cols(0, "2:4", sep="", outsep="!") == [
        "H",
        "l!l",
    ]  # Non-empty outsep

    assert text.cols(0, "2:4", "1:4", sep="", outsep=".") == [
        "H",
        "l.l",
        "e.l.l",
    ]  # Overlapping columns

    # Character-by-character splitting for DNA sequence
    dna = EStr("ATCGTA")
    assert dna.cols(sep="") == ["A", "T", "C", "G", "T", "A"]
    assert dna.cols("1,3,5", sep="") == "T G A"
    assert dna.cols("0:3", sep="", outsep="") == "ATC"

    # Multiple arguments with character splitting
    text = EStr("Hello world!")
    assert text.cols("0", "1:5", sep="") == ["H", "e l l o"]
    assert text.cols("0:4", "6:9", sep="", outsep="-") == ["H-e-l-l", "w-o-r"]

    # Unicode characters
    emoji = EStr("👋🌍")
    assert emoji.cols(sep="") == ["👋", "🌍"]
    assert emoji.cols("0", sep="") == "👋"


def test_estr_cols_with_regex():
    s = EStr("This2334is7453a654test232with232regex")
    assert s.cols("1:5", sep=re.compile(r"\d+")) == "is a test with"


def test_extract_json_valid_json_object():
    assert extract_json('{"name": "John", "age": 30}') == '{"name": "John", "age": 30}'


def test_extract_json_valid_json_array():
    assert extract_json("[1, 2, 3, 4]") == "[1, 2, 3, 4]"


def test_extract_json_nested_json():
    assert (
        extract_json('{"person": {"name": "John", "age": 30}, "city": "New York"}')
        == '{"person": {"name": "John", "age": 30}, "city": "New York"}'
    )


def test_extract_json_json_with_text_before():
    assert (
        extract_json('Hello world {"name": "John", "age": 30}')
        == '{"name": "John", "age": 30}'
    )


def test_extract_json_json_with_text_after():
    assert (
        extract_json('{"name": "John", "age": 30} and more text')
        == '{"name": "John", "age": 30}'
    )


def test_extract_json_invalid_json():
    with pytest.raises(ValueError):
        extract_json('{"name": "John" "age": 30}')


def test_extract_json_no_json():
    with pytest.raises(ValueError):
        extract_json("Just a plain text without JSON")


def test_extract_json_multiple_json_objects():
    assert extract_json('{"name": "John"} {"age": 30}') == '{"name": "John"}'


def test_extract_json_empty_string():
    with pytest.raises(ValueError):
        extract_json("")
